import warnings
from typing import cast

from starlette.middleware import Middleware
from starlette.middleware.authentication import AuthenticationMiddleware

from starlette_proxy_router.exceptions import ProxyAuthenficationError
from starlette_proxy_router.parser import EndpointArgs


def check_auth_midlleware(
    middlewares: list[Middleware], endpoints_args: list[EndpointArgs]
) -> None:
    auth_midlleware = any(
        issubclass(cast(type, middleware.cls), AuthenticationMiddleware)
        for middleware in middlewares
    )
    auth_endpoints = any(args.is_auth for args in endpoints_args)

    if auth_midlleware and not auth_endpoints:
        warnings.warn(
            "Existing AuthenticationMiddleware is not used in proxy routes",
            DeprecationWarning,
            stacklevel=2,
        )
    elif not auth_midlleware and auth_endpoints:
        raise ProxyAuthenficationError(
            "The application does not have AuthenticationMiddleware implemented or added"
        )



from starlette.routing import Route

from starlette_proxy_router.compile import RouteCompilier
from starlette_proxy_router.parser import EndpointArgs


def create_routes(endpoints_args: list[EndpointArgs]) -> list[Route]:
    compilier = RouteCompilier()
    routes = []
    for args in endpoints_args:
        route = compilier.compile_route(args)
        routes.append(route)
    return routes



from starlette_proxy_router.parser import EndpointArgs, Parser
from starlette_proxy_router.types import AuthType
from starlette_proxy_router.validate import MethodHTTP
from tests.constants import PATH_FILE_CONFIG


def test_parser() -> None:
    parser = Parser(PATH_FILE_CONFIG)
    args_endpoints = parser.parse_configs()
    assert len(args_endpoints) == 6
    assert args_endpoints[0] == EndpointArgs(
        path="/api/v1/proxy",
        target="/api/v1/target",
        method=MethodHTTP.GET,
        host="http://localhost:8000/",
        service_name="test_service_1",
        is_auth=True,
        timeout=None,
        auth_type=AuthType.QUERY_PARAMS,
        auth_fields=["user_id"],
    )
    assert args_endpoints[1] == EndpointArgs(
        path="/api/v1/proxy",
        target="/api/v1/target",
        method=MethodHTTP.PUT,
        host="http://localhost:8000/",
        service_name="test_service_1",
        is_auth=False,
        timeout=None,
        auth_type=AuthType.QUERY_PARAMS,
        auth_fields=["user_id"],
    )
    assert args_endpoints[4] == EndpointArgs(
        path="/api/v1/create-object",
        target="/api/v1/create-object-target",
        method=MethodHTTP.GET,
        host="http://localhost:8000/",
        service_name="test_service_2",
        is_auth=True,
        timeout=10,
        auth_type=AuthType.QUERY_PARAMS,
        auth_fields=["user_id"],
    )
    assert args_endpoints[5] == EndpointArgs(
        path="/",
        target="/",
        method=MethodHTTP.GET,
        host="http://localhost:8000/",
        service_name="test_service_2",
        is_auth=True,
        timeout=None,
        auth_type=AuthType.QUERY_PARAMS,
        auth_fields=["user_id"],
    )



class BaseProxyRouterError(Exception):
    pass


class ProxyAuthenficationError(BaseProxyRouterError):
    pass


class AttributeAuthUserError(BaseProxyRouterError):
    pass


class ConfigTypeError(BaseProxyRouterError):
    pass


class HostNotFoundError(BaseProxyRouterError):
    pass



from starlette_proxy_router.compile import RouteCompilier, _build_target_url


def test_build_target_url() -> None:
    target_url = _build_target_url("http://localhost:8000/", "/api/v1/test", {})
    assert target_url == "http://localhost:8000/api/v1/test"
    target_url = _build_target_url(
        "http://localhost:8000/", target="/api/v1/test/{user_id}", path_params={"user_id": "1"}
    )
    assert target_url == "http://localhost:8000/api/v1/test/1"


def test_build_func_name() -> None:
    compilier = RouteCompilier()
    func_name = compilier._build_func_name("/api/v1/test")
    assert func_name == "api_v1_test"

    func_name = compilier._build_func_name("/")
    assert func_name == "homepage"

    func_name = compilier._build_func_name("/api/v1/{user_id}/test")
    assert func_name == "api_v1_test"



import json
import os
from typing import NamedTuple

from starlette_proxy_router.exceptions import HostNotFoundError
from starlette_proxy_router.types import AuthType, PathConfigs, PathURL, ProxyPathURL
from starlette_proxy_router.validate import ConfigValidate, HostValidate, MethodHTTP


class EndpointArgs(NamedTuple):
    path: PathURL
    target: ProxyPathURL
    method: MethodHTTP
    host: str
    service_name: str
    is_auth: bool = False
    timeout: int = 10
    auth_type: AuthType | None = None
    auth_fields: list[str] | None = None


def _check_exists_path(path_configs: PathConfigs) -> None:
    if isinstance(path_configs, list):
        for path in path_configs:
            if not os.path.exists(path):
                raise FileNotFoundError(f'The specified path does not exist: {path}')
    else:
        if not os.path.exists(path_configs):
            raise FileNotFoundError(f'The specified path does not exist: {path_configs}')


class Parser:
    __slots__ = ("_path_configs",)

    def __init__(self, path_configs: PathConfigs) -> None:
        _check_exists_path(path_configs)
        self._path_configs = path_configs

    def parse_configs(self) -> list[EndpointArgs]:
        if isinstance(self._path_configs, str):
            if os.path.isdir(self._path_configs):
                path_configs = [
                    os.path.join(self._path_configs, path)
                    for path in os.listdir(self._path_configs)
                    if os.path.isfile(os.path.join(self._path_configs, path))
                ]
            else:
                path_configs = [self._path_configs]

        path_configs = [path for path in path_configs if path.endswith('json')]
        if not path_configs:
            raise ValueError("No JSON configs found in specified paths")

        result = []
        for path in path_configs:
            with open(path) as file:
                content = json.load(file)
                data = self._validate_config(content)
                endpoint_args = self._parse(data)
                result.extend(endpoint_args)
        return result

    def _validate_config(self, data: dict) -> ConfigValidate:
        return ConfigValidate.model_validate(data)

    def _validate_and_get_host(self, service_name: str) -> str:
        host = os.environ.get(service_name.upper())
        if host:
            host_validate = HostValidate.model_validate({'host': host})
            return str(host_validate.host)
        raise HostNotFoundError(f"Host for service '{service_name}' not found in env")

    def _parse(self, config_validate: ConfigValidate) -> list[EndpointArgs]:
        args_endpoints = []
        for service in config_validate.services:
            for endpoint in service.endpoints:
                args = EndpointArgs(
                    path=endpoint.path,
                    target=endpoint.target if endpoint.target else endpoint.path,
                    method=endpoint.method or MethodHTTP.GET,
                    is_auth=(
                        endpoint.is_auth
                        if endpoint.is_auth is not None
                        else config_validate.authenfication.default_on_endpoint
                    ),
                    auth_type=config_validate.authenfication.default_type,
                    auth_fields=config_validate.authenfication.fields,
                    service_name=service.service,
                    host=self._validate_and_get_host(service.service),
                    timeout=endpoint.timeout,
                )
                args_endpoints.append(args)
        return args_endpoints



from starlette.routing import Route

from starlette_proxy_router.parser import Parser
from starlette_proxy_router.routing import create_routes
from tests.constants import PATH_FILE_CONFIG


def test_create_routes() -> None:
    parser = Parser(PATH_FILE_CONFIG)
    endpoints_args = parser.parse_configs()
    routes = create_routes(endpoints_args)
    assert isinstance(routes[0], Route) is True
    assert routes[0].path == "/api/v1/proxy"



import json
import linecache
import re
from collections.abc import Mapping
from typing import NamedTuple

from httpx import AsyncClient
from starlette.authentication import UnauthenticatedUser
from starlette.datastructures import Headers
from starlette.exceptions import HTTPException
from starlette.requests import Request
from starlette.responses import Response
from starlette.routing import Route

from starlette_proxy_router.exceptions import AttributeAuthUserError
from starlette_proxy_router.parser import EndpointArgs, MethodHTTP
from starlette_proxy_router.types import AuthType


class DataRequest(NamedTuple):
    headers: Headers
    query_params: Mapping[str, str]
    path_params: Mapping[str, str]
    body: bytes


async def _parse_request(request: Request) -> DataRequest:
    return DataRequest(
        headers=request.scope["headers"],
        query_params=request.query_params,
        body=await request.body(),
        path_params=request.path_params,
    )


def _build_target_url(host: str, target: str, path_params: Mapping[str, str]) -> str:
    target = target.format_map(path_params)
    return f"{host}{target[1:]}"


_AUTH_CODE_LINES = """
    is_auth = {is_auth}
    auth_type = '{auth_type}'
    auth_fields = {auth_fields}

    query_params = {{k: v for k, v in data_request.query_params.items()}}
    exclude_keys = [b'content-length', b'host']
    headers = {{k.decode(): v.decode() for k, v in data_request.headers if k not in exclude_keys}}

    if is_auth:
        if isinstance(request.user, UnauthenticatedUser):
            raise HTTPException(status_code=401, detail='User is not authenticated')
        if auth_type == AuthType.QUERY_PARAMS:
            for field in auth_fields:
                try:
                    query_params[field] = request.user.__dict__[field]
                except AttributeError as e:
                    raise AttributeAuthUserError(field)
        else:
            for field in auth_fields:
                try:
                    headers[field.replace('_', '-').capitalize()] = request.user.__dict__[field]
                except AttributeError as e:
                    raise AttributeAuthUserError(field)
"""
_HTTP_PROXY_ENDPOINT_IS_BODY = """
async def {func_name}(request):
    data_request = await _parse_request(request)
    target_url = _build_target_url('{host}', '{target}', data_request.path_params)

    {auth_code_lines}

    try:
        body = json.loads(data_request.body)
    except json.decoder.JSONDecodeError:
        body = None

    async with AsyncClient() as ac:
        res = await ac.{method}(
            target_url,
            params=query_params,
            json=body,
            timeout={timeout},
            headers=headers
        )
        return Response(
            content=res.content,
            status_code=res.status_code,
            headers=res.headers,
        )
"""
_HTTP_PROXY_ENDPOINT_NOT_BODY = """
async def {func_name}(request):
    data_request = await _parse_request(request)
    target_url = _build_target_url('{host}', '{target}', data_request.path_params)

    {auth_code_lines}

    async with AsyncClient() as ac:
        res = await ac.{method}(
            target_url,
            params=query_params,
            timeout={timeout},
            headers=headers
        )
        return Response(
            content=res.content,
            status_code=res.status_code,
            headers=res.headers,
        )
"""
_METHOD_BODIES = {
    MethodHTTP.GET: _HTTP_PROXY_ENDPOINT_NOT_BODY,
    MethodHTTP.POST: _HTTP_PROXY_ENDPOINT_IS_BODY,
    MethodHTTP.PUT: _HTTP_PROXY_ENDPOINT_IS_BODY,
    MethodHTTP.DELETE: _HTTP_PROXY_ENDPOINT_NOT_BODY,
    MethodHTTP.PATCH: _HTTP_PROXY_ENDPOINT_IS_BODY
}


class RouteCompilier:
    __slots__ = ()

    def compile_route(self, endpoint_args: EndpointArgs) -> Route:
        func_name = self._build_func_name(endpoint_args.path)
        body_auth_code = _AUTH_CODE_LINES.format_map({
            "is_auth": endpoint_args.is_auth,
            "auth_type": endpoint_args.auth_type,
            "auth_fields": endpoint_args.auth_fields,
        })
        body_http_endpoint = _METHOD_BODIES[endpoint_args.method]
        body_http_endpoint = body_http_endpoint.format_map({
            "auth_code_lines": body_auth_code,
            "func_name": func_name,
            "method": endpoint_args.method.value,
            "target": endpoint_args.target,
            "host": endpoint_args.host,
            "timeout": endpoint_args.timeout,
        })
        global_vars = {
            'json': json,
            "AsyncClient": AsyncClient,
            "Response": Response,
            "AuthType": AuthType,
            "AttributeAuthUserError": AttributeAuthUserError,
            "HTTPException": HTTPException,
            "UnauthenticatedUser": UnauthenticatedUser,
            "_parse_request": _parse_request,
            "_build_target_url": _build_target_url,
        }
        file_name = f"{endpoint_args.service_name}_route"
        lines = body_http_endpoint.splitlines(keepends=True)
        linecache.cache[file_name] = (
            len(body_http_endpoint),
            None,
            lines,
            file_name,
        )
        compiled = compile(body_http_endpoint, file_name, "exec")
        exec(compiled, global_vars)

        return Route(
            endpoint_args.path,
            global_vars[func_name],  # type: ignore
            methods=[endpoint_args.method],
        )

    def _build_func_name(self, path: str) -> str:
        sub_paths = re.sub(r"/{[^{}]+}", "", path)
        sub_paths_list = re.split(r"[-_/]", sub_paths)
        result = "_".join(sub_paths_list[1:])

        if not result:
            return "homepage"
        return result



from enum import StrEnum
from typing import TypeAlias

PathURL: TypeAlias = str
ProxyPathURL: TypeAlias = str

PathFileConfig: TypeAlias = str
PathDirectoryConfig: TypeAlias = str
PathConfigs: TypeAlias = PathFileConfig | PathDirectoryConfig | list[PathFileConfig]


class AuthType(StrEnum):
    QUERY_PARAMS = "query_params"
    HEADERS = "headers"



from starlette.applications import Starlette

from starlette_proxy_router.authenfication import check_auth_midlleware
from starlette_proxy_router.parser import Parser
from starlette_proxy_router.routing import create_routes
from starlette_proxy_router.types import PathConfigs


def create_proxy_router(app: Starlette, path_configs: PathConfigs = "config.json") -> None:
    parser = Parser(path_configs)
    endpoints_args = parser.parse_configs()
    check_auth_midlleware(app.user_middleware, endpoints_args)
    proxy_routes = create_routes(endpoints_args)
    app.router.routes.extend(proxy_routes)  # type: ignore



import base64
import binascii

from starlette.authentication import (
    AuthCredentials,
    AuthenticationBackend,
    AuthenticationError,
    BaseUser,
)
from starlette.middleware import Middleware
from starlette.middleware.authentication import AuthenticationMiddleware
from starlette.requests import HTTPConnection


class User(BaseUser):
    def __init__(self, user_id: str) -> None:
        self.user_id = user_id

    @property
    def is_authenticated(self) -> bool:
        return True


class BasicAuthBackend(AuthenticationBackend):
    async def authenticate(self, conn: HTTPConnection) -> tuple[AuthCredentials, User] | None:
        if "Authorization" not in conn.headers:
            return None

        auth = conn.headers["Authorization"]
        try:
            scheme, credentials = auth.split()
            if scheme.lower() != "basic":
                return None
            decoded = base64.b64decode(credentials).decode("ascii")
        except (ValueError, UnicodeDecodeError, binascii.Error):
            raise AuthenticationError("Invalid basic auth credentials")

        user_id, _, _ = decoded.partition(":")
        return AuthCredentials(["authenticated"]), User(user_id)


def create_auth_middleware() -> list[Middleware]:
    return [Middleware(AuthenticationMiddleware, backend=BasicAuthBackend())]



import base64
from collections.abc import AsyncGenerator, Generator
from multiprocessing import Process

import pytest
import pytest_asyncio
import uvicorn
from httpx import ASGITransport, AsyncClient
from starlette.applications import Starlette
from starlette.requests import Request
from starlette.responses import PlainTextResponse
from starlette.routing import Route

from starlette_proxy_router import create_proxy_router
from tests.constants import PATH_DIR_CONFIG
from tests.utils.auth import create_auth_middleware


def homepage(request: Request) -> PlainTextResponse:
    user_id = request.query_params.get("user_id", "")
    return PlainTextResponse(f"Hello, {user_id}!")


def homepage_not_user_id(request: Request) -> PlainTextResponse:
    return PlainTextResponse("Hello, world!")


def create_object(request: Request) -> PlainTextResponse:
    return PlainTextResponse("Object created", status_code=201)


def get_branch_id(request: Request) -> PlainTextResponse:
    return PlainTextResponse(f"Branch: {request.path_params['branch_id']}")


def run_target_app() -> None:
    base_routes = [
        Route("/api/v1/target", homepage, methods=["GET"]),
        Route("/api/v1/target", homepage_not_user_id, methods=["PUT"]),
        Route("/api/v1/target", create_object, methods=["POST"]),
        Route("/api/v1/target", homepage, methods=["DELETE"]),
        Route("/api/v1/create-object-target", homepage, methods=["GET"]),
        Route("/", homepage, methods=["GET"]),
        Route("/additional-test", homepage, methods=["GET"]),
        Route("/additional-test/{branch_id}", get_branch_id, methods=['GET'])
    ]
    app = Starlette(routes=base_routes)
    uvicorn.run(app)


@pytest.fixture(scope="module", autouse=True)
def run_target_server() -> Generator[None]:
    proc = Process(target=run_target_app, args=(), daemon=True)
    proc.start()
    yield
    proc.kill()


@pytest.fixture(scope="module")
def proxy_app() -> Starlette:
    middleware = create_auth_middleware()
    app = Starlette(middleware=middleware)
    create_proxy_router(app, path_configs=PATH_DIR_CONFIG)
    return app


@pytest.fixture
def headers() -> dict[str, str]:
    credentials = "test_user_id:test_password"
    encoded_credentials = base64.b64encode(credentials.encode("ascii")).decode("ascii")
    return {"Authorization": f"Basic {encoded_credentials}"}


@pytest_asyncio.fixture
async def client(proxy_app: Starlette) -> AsyncGenerator[AsyncClient]:
    async with AsyncClient(
        transport=ASGITransport(app=proxy_app), base_url="http://test"
    ) as client:
        yield client


@pytest.mark.parametrize("url", ["/api/v1/proxy", "/api/v1/create-object", "/", "/additional-test"])
async def test_method_get_proxy(client: AsyncClient, url: str, headers: dict[str, str]) -> None:
    res = await client.get(url, headers=headers)
    assert res.status_code == 200
    assert res.text == "Hello, test_user_id!"


async def test_methods_proxy(client: AsyncClient, headers: dict[str, str]) -> None:
    res = await client.post("/api/v1/proxy", headers=headers)
    assert res.status_code == 201
    assert res.text == "Object created"

    res = await client.put("/api/v1/proxy", headers=headers)
    assert res.status_code == 200
    assert res.text == "Hello, world!"

    res = await client.delete("/api/v1/proxy", headers=headers)
    assert res.status_code == 200
    assert res.text == "Hello, test_user_id!"


async def test_path_params(client: AsyncClient, headers: dict[str, str]) -> None:
    res = await client.get("/additional-test/1", headers=headers)
    assert res.status_code == 200
    assert res.text == "Branch: 1"



from enum import StrEnum

from pydantic import BaseModel, HttpUrl, field_validator

from starlette_proxy_router.types import AuthType, PathURL, ProxyPathURL


class MethodHTTP(StrEnum):
    GET = "get"
    POST = "post"
    DELETE = "delete"
    PUT = "put"
    PATCH = "patch"


class Endpoint(BaseModel):
    path: PathURL
    target: ProxyPathURL | None = None
    method: MethodHTTP | None = None
    is_auth: bool | None = None
    timeout: int | None = 10

    @field_validator("path", "target", mode="before")
    @classmethod
    def startswith(cls, value: PathURL | ProxyPathURL) -> PathURL | ProxyPathURL:
        if not value.startswith("/"):
            return f"/{value}"
        return value


class Service(BaseModel):
    service: str
    endpoints: list[Endpoint]


class ConfigAuthenfication(BaseModel):
    default_type: AuthType | None = None
    fields: list[str] | None = None
    default_on_endpoint: bool = False


class ConfigValidate(BaseModel):
    services: list[Service]
    authenfication: ConfigAuthenfication = ConfigAuthenfication()


class HostValidate(BaseModel):
    host: HttpUrl



import pytest

from starlette_proxy_router.authenfication import check_auth_midlleware
from starlette_proxy_router.exceptions import ProxyAuthenficationError
from starlette_proxy_router.parser import EndpointArgs, Parser
from tests.constants import PATH_FILE_CONFIG
from tests.utils.auth import create_auth_middleware


@pytest.fixture
def endpoints_args() -> list[EndpointArgs]:
    parser = Parser(PATH_FILE_CONFIG)
    return parser.parse_configs()


def test_check_auth_midlleware(endpoints_args: list[EndpointArgs]) -> None:
    auth_midlleware = create_auth_middleware()

    assert check_auth_midlleware(auth_midlleware, endpoints_args) is None  # type: ignore

    with pytest.raises(ProxyAuthenficationError):
        check_auth_midlleware([], endpoints_args)

    endpoints_args = [
        EndpointArgs(
            path=args.path,
            target=args.target,
            method=args.method,
            host=args.host,
            service_name=args.service_name,
            is_auth=False,
        )
        for args in endpoints_args
    ]
    with pytest.deprecated_call():
        check_auth_midlleware(auth_midlleware, endpoints_args)
